/*-
 * Copyright 2022 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.quality;

import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.citygml4j.core.ade.ADEException;
import org.citygml4j.core.ade.ADERegistry;
import org.citygml4j.core.model.CityGMLVersion;
import org.citygml4j.core.model.building.Building;
import org.citygml4j.core.model.core.AbstractCityObjectProperty;
import org.citygml4j.core.model.core.AbstractFeatureProperty;
import org.citygml4j.core.model.core.CityModel;
import org.citygml4j.core.util.CityGMLConstants;
import org.citygml4j.xml.CityGMLContext;
import org.citygml4j.xml.CityGMLContextException;
import org.citygml4j.xml.module.citygml.CoreModule;
import org.citygml4j.xml.schema.CityGMLSchemaHandler;
import org.citygml4j.xml.writer.CityGMLOutputFactory;
import org.citygml4j.xml.writer.CityGMLWriteException;
import org.citygml4j.xml.writer.CityGMLWriter;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xmlobjects.gml.model.base.Reference;
import org.xmlobjects.gml.model.geometry.DirectPosition;
import org.xmlobjects.gml.model.measures.Angle;
import org.xmlobjects.gml.model.measures.Length;
import org.xmlobjects.schema.SchemaHandlerException;

import de.hft.stuttgart.quality.model.enums.ErrorId;
import de.hft.stuttgart.quality.model.enums.RequirementId;
import de.hft.stuttgart.quality.model.enums.ResultType;
import de.hft.stuttgart.quality.model.enums.RingSelfIntType;
import de.hft.stuttgart.quality.model.enums.TopLevelFeatureType;
import de.hft.stuttgart.quality.model.properties.AbstractErrorProperty;
import de.hft.stuttgart.quality.model.properties.CheckingProperty;
import de.hft.stuttgart.quality.model.properties.EdgeListProperty;
import de.hft.stuttgart.quality.model.properties.EdgeProperty;
import de.hft.stuttgart.quality.model.properties.ErrorProperty;
import de.hft.stuttgart.quality.model.properties.FeatureStatisticsProperty;
import de.hft.stuttgart.quality.model.properties.FilterProperty;
import de.hft.stuttgart.quality.model.properties.GlobalParametersProperty;
import de.hft.stuttgart.quality.model.properties.ParameterProperty;
import de.hft.stuttgart.quality.model.properties.PolygonIdListProperty;
import de.hft.stuttgart.quality.model.properties.RequirementProperty;
import de.hft.stuttgart.quality.model.properties.StatisticsProperty;
import de.hft.stuttgart.quality.model.properties.ValidationPlanProperty;
import de.hft.stuttgart.quality.model.properties.ValidationResultProperty;
import de.hft.stuttgart.quality.model.types.AllPolygonsOrientedWrongError;
import de.hft.stuttgart.quality.model.types.Checking;
import de.hft.stuttgart.quality.model.types.CityObjectProperties;
import de.hft.stuttgart.quality.model.types.ConsecutivePointsSameError;
import de.hft.stuttgart.quality.model.types.Edge;
import de.hft.stuttgart.quality.model.types.EdgeList;
import de.hft.stuttgart.quality.model.types.FeatureStatistics;
import de.hft.stuttgart.quality.model.types.Filter;
import de.hft.stuttgart.quality.model.types.GlobalParameters;
import de.hft.stuttgart.quality.model.types.HoleOutsideError;
import de.hft.stuttgart.quality.model.types.InnerRingsNestedError;
import de.hft.stuttgart.quality.model.types.InteriorDisconnectedError;
import de.hft.stuttgart.quality.model.types.IntersectingRingsError;
import de.hft.stuttgart.quality.model.types.MultipleComponentsError;
import de.hft.stuttgart.quality.model.types.NonManifoldEdgeError;
import de.hft.stuttgart.quality.model.types.NonManifoldVertexError;
import de.hft.stuttgart.quality.model.types.OrientationRingsSameError;
import de.hft.stuttgart.quality.model.types.Parameter;
import de.hft.stuttgart.quality.model.types.PlanarDistancePlaneError;
import de.hft.stuttgart.quality.model.types.PlanarNormalsDeviationError;
import de.hft.stuttgart.quality.model.types.PolygonIdList;
import de.hft.stuttgart.quality.model.types.PolygonWrongOrientationError;
import de.hft.stuttgart.quality.model.types.Requirement;
import de.hft.stuttgart.quality.model.types.RingNotClosedError;
import de.hft.stuttgart.quality.model.types.RingSelfIntersectionError;
import de.hft.stuttgart.quality.model.types.RingTooFewPointsError;
import de.hft.stuttgart.quality.model.types.SemanticAttributeMissingError;
import de.hft.stuttgart.quality.model.types.SemanticAttributeWrongValueError;
import de.hft.stuttgart.quality.model.types.SolidNotClosedError;
import de.hft.stuttgart.quality.model.types.SolidSelfIntersectionError;
import de.hft.stuttgart.quality.model.types.SolidTooFewPolygonsError;
import de.hft.stuttgart.quality.model.types.Statistics;
import de.hft.stuttgart.quality.model.types.Validation;
import de.hft.stuttgart.quality.model.types.ValidationPlan;
import de.hft.stuttgart.quality.model.types.ValidationResult;

class QualityAdeTests {

	private static final String CITY_OBJECT_MEMBER = "cityObjectMember";
	private static List<QName> chunkProperties = new ArrayList<>();

	static {
		chunkProperties.add(new QName(CityGMLConstants.CITYGML_1_0_CORE_NAMESPACE, CITY_OBJECT_MEMBER));
		chunkProperties.add(new QName(CityGMLConstants.CITYGML_2_0_CORE_NAMESPACE, CITY_OBJECT_MEMBER));
		chunkProperties.add(new QName(CityGMLConstants.CITYGML_3_0_CORE_NAMESPACE, CITY_OBJECT_MEMBER));
	}

	@BeforeAll
	static void setUsLocale() throws ADEException {
		Locale.setDefault(Locale.US);
		ADERegistry.getInstance().loadADE(new QualityADEContext());
	}

	@Test
	void testAllPolygonsOrientedWrong() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		AllPolygonsOrientedWrongError err = new AllPolygonsOrientedWrongError();
		err.setGeometryId("geomId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testConsecutivePointsSameError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		ConsecutivePointsSameError err = new ConsecutivePointsSameError();
		err.setLinearRingId("ringId");
		err.setVertex1(new DirectPosition(1, 2, 3));
		err.setVertex2(new DirectPosition(5, 6, 7));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testHoleOutsideError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		HoleOutsideError err = new HoleOutsideError();
		err.setLinearRingId("ringId");
		err.setPolygonId("polyId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testInnerRingsNestedError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		InnerRingsNestedError err = new InnerRingsNestedError();
		err.setLinearRingId1("ringId1");
		err.setLinearRingId2("ringId2");
		err.setPolygonId("polyId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testInteriorDisconnectedError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		InteriorDisconnectedError err = new InteriorDisconnectedError();
		err.setPolygonId("polyId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testIntersectingRingsError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		IntersectingRingsError err = new IntersectingRingsError();
		err.setLinearRingId1("ringId1");
		err.setLinearRingId2("ringId2");
		err.setPolygonId("polyId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testMultipleComponentsError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		MultipleComponentsError err = new MultipleComponentsError();
		err.setGeometryId("geomId");
		PolygonIdList polygons = new PolygonIdList();
		polygons.getPolygonIds().add("test1");
		polygons.getPolygonIds().add("test2");
		err.getComponents().add(new PolygonIdListProperty(polygons));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testNonManifoldEdgeError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		NonManifoldEdgeError err = new NonManifoldEdgeError();
		err.setGeometryId("geomId");
		EdgeList edges = new EdgeList();
		Edge edge = new Edge();
		edge.setFrom(new DirectPosition(1, 2, 3));
		edge.setTo(new DirectPosition(5, 6, 7));
		edges.getEdges().add(new EdgeProperty(edge));
		err.setEdges(new EdgeListProperty(edges));

		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testNonManifoldVertexError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		NonManifoldVertexError err = new NonManifoldVertexError();
		err.setGeometryId("geomId");
		err.setVertex(new DirectPosition(1, 2, 3));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testOrientationRingsSameError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		OrientationRingsSameError err = new OrientationRingsSameError();
		err.setLinearRing("ring1");
		err.setPolygonId("polyId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testPlanarDistancePlaneError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		PlanarDistancePlaneError err = new PlanarDistancePlaneError();
		err.setPolygonId("polyId");
		err.setDistance(new Length(5d, "m"));
		err.setVertex(new DirectPosition(1, 2, 3));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testPlanarNormalsDeviationError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		PlanarNormalsDeviationError err = new PlanarNormalsDeviationError();
		err.setPolygonId("polyId");
		err.setDeviation(new Angle(5d, "deg"));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testPolygonWrongOrientationError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		PolygonWrongOrientationError err = new PolygonWrongOrientationError();
		err.setGeometryId("geomId");
		EdgeList edges = new EdgeList();
		Edge edge = new Edge();
		edge.setFrom(new DirectPosition(1, 2, 3));
		edge.setTo(new DirectPosition(5, 6, 7));
		edges.getEdges().add(new EdgeProperty(edge));
		err.setEdges(new EdgeListProperty(edges));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testRingNotClosedError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		RingNotClosedError err = new RingNotClosedError();
		err.setLinearRingId("ringId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testRingSelfIntersectionError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		RingSelfIntersectionError err = new RingSelfIntersectionError();
		err.setLinearRingId("ringId");
		Edge edge = new Edge();
		edge.setFrom(new DirectPosition(1, 2, 3));
		edge.setTo(new DirectPosition(5, 6, 7));
		err.setEdge1(new EdgeProperty(edge));
		err.setEdge2(new EdgeProperty(edge));
		err.setType(RingSelfIntType.EDGE_INTERSECTION);
		err.setVertex1(new DirectPosition(1, 2, 3));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testRingTooFewPointsError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		RingTooFewPointsError err = new RingTooFewPointsError();
		err.setLinearRingId("ringId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testSemanticAttributeMissingError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		SemanticAttributeMissingError err = new SemanticAttributeMissingError();
		err.setAttributeName("attribute");
		err.setChildId("child");
		err.setGeneric(true);
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testSemanticAttributeWrongValueError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		SemanticAttributeWrongValueError err = new SemanticAttributeWrongValueError();
		err.setAttributeName("attribute");
		err.setChildId("child");
		err.setGeneric(true);
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testSolidNotClosedError() throws ADEException, CityGMLContextException, CityGMLWriteException, SAXException,
			IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		SolidNotClosedError err = new SolidNotClosedError();
		err.setGeometryId("geomId");
		EdgeList edges = new EdgeList();
		Edge edge = new Edge();
		edge.setFrom(new DirectPosition(1, 2, 3));
		edge.setTo(new DirectPosition(5, 6, 7));
		edges.getEdges().add(new EdgeProperty(edge));
		err.setEdges(new EdgeListProperty(edges));
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testSolidSelfIntersectionError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		SolidSelfIntersectionError err = new SolidSelfIntersectionError();
		err.setGeometryId("geomId");
		err.setPolygonId1("poly1");
		err.setPolygonId2("poly2");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	@Test
	void testSolidTooFewPolygonsError() throws ADEException, CityGMLContextException, CityGMLWriteException,
			SAXException, IOException, SchemaHandlerException {
		CityModel model = new CityModel();
		ValidationResult res = fillCityModel(model);

		SolidTooFewPolygonsError err = new SolidTooFewPolygonsError();
		err.setGeometryId("geomId");
		res.getErrors().add(new AbstractErrorProperty(err));

		byte[] buf = writeModel(model);
		assertTrue(validate(new ByteArrayInputStream(buf)));
	}

	private boolean validate(InputStream stream)
			throws SAXException, IOException, CityGMLContextException, SchemaHandlerException {

		CityGMLContext context = CityGMLContext.newInstance();

		CityGMLSchemaHandler schemaHandler = context.getDefaultSchemaHandler();
		Source[] schemas = schemaHandler.getSchemas();

		SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		Schema schema = schemaFactory.newSchema(schemas);

		Validator validator = schema.newValidator();
		boolean[] foundErrors = new boolean[1];
		foundErrors[0] = false;

		validator.setErrorHandler(new ErrorHandler() {
			@Override
			public void error(SAXParseException exception) {
				String message = "[" + exception.getLineNumber() + ", " + exception.getColumnNumber() + "] "
						+ exception.getMessage();
				System.out.println(message);
				foundErrors[0] = true;
			}

			@Override
			public void warning(SAXParseException exception) {
				error(exception);
			}

			@Override
			public void fatalError(SAXParseException exception) {
				error(exception);
			}
		});

		validator.validate(new StreamSource(stream));
		return !foundErrors[0];
	}

	private byte[] writeModel(CityModel model) throws ADEException, CityGMLContextException, CityGMLWriteException {

		CityGMLContext context = CityGMLContext.newInstance();
		CityGMLVersion version = CityGMLVersion.v2_0;
		CityGMLOutputFactory out = context.createCityGMLOutputFactory(version);
		ByteArrayOutputStream outStream = new ByteArrayOutputStream();
		try (CityGMLWriter writer = out.createCityGMLWriter(outStream, StandardCharsets.UTF_8.name())) {
			writer.withIndent("  ")
					.withSchemaLocation(QualityADEModule.NAMESPACE_URI,
							getClass().getResource("/qualityAde.xsd").toString())
					.withDefaultPrefixes().withDefaultNamespace(CoreModule.of(version).getNamespaceURI()).write(model);
		}
		return outStream.toByteArray();
	}

	private ValidationResult fillCityModel(CityModel model) {
		Validation v = new Validation();
		v.setId("planId");
		model.getFeatureMembers().add(new AbstractFeatureProperty(v));
		v.setValidationSoftware("testSoftware");
		v.setValidationDate(ZonedDateTime.now());
		ValidationPlan plan = new ValidationPlan();

		Statistics stats = new Statistics();
		FeatureStatistics fStats = new FeatureStatistics();
		fStats.setNumChecked(5);
		fStats.setNumErrors(2);
		stats.setNumErrorBridgeObjects(new FeatureStatisticsProperty(fStats));
		v.setStatistics(new StatisticsProperty(stats));
		
		de.hft.stuttgart.quality.model.types.Error err = new de.hft.stuttgart.quality.model.types.Error();
		err.setName(ErrorId.GE_P_HOLE_OUTSIDE);
		err.setOccurrences(1);
		stats.getErrors().add(new ErrorProperty(err));

		Requirement req = new Requirement();
		req.setEnabled(true);
		req.setRequirementType(RequirementId.R_GE_P_HOLE_OUTSIDE);
		plan.getRequirements().add(new RequirementProperty(req));

		Checking checking = new Checking();
		checking.setFeatureType(TopLevelFeatureType.BRIDGE);

		Filter filter = new Filter();
		filter.getChecking().add(new CheckingProperty(checking));
		plan.setFilter(new FilterProperty(filter));

		Parameter param = new Parameter();
		param.setName("test");
		param.setUom("uom");
		param.setValue("5");
		GlobalParameters globalParams = new GlobalParameters();
		globalParams.getParameters().add(new ParameterProperty(param));
		plan.setGlobalParameters(new GlobalParametersProperty(globalParams));

		v.setValidationPlan(new ValidationPlanProperty(plan));

		Building b = new Building();
		b.setId("testId");
		model.getCityObjectMembers().add(new AbstractCityObjectProperty(b));

		CityObjectProperties props = new CityObjectProperties();
		b.addADEProperty(props);
		ValidationResult res = new ValidationResult();
		props.setValidationResult(new ValidationResultProperty(res));
		res.setResultType(ResultType.ERROR);
		res.setValidationPlanID(new Reference(v));
		return res;
	}
}
